from aloft.chart_config import get_charts_to_use
from aloft import cluster


def get_create_vpc_args(arguments):
    return get_vpc_args(arguments)


def get_delete_vpc_args(arguments):
    return get_vpc_args(arguments)


def get_vpc_args(arguments):
    vpc_id = arguments['<vpc_id>']
    return vpc_id


def get_apply_chart_args(arguments):
    return get_chart_args_with_options(arguments)


def get_delete_chart_args(arguments):
    return get_chart_args_with_options(arguments)


def get_lock_volumes_args(arguments):
    return get_chart_args(arguments)


def get_unlock_volumes_args(arguments):
    return get_chart_args(arguments)


def get_chart_args_with_options(arguments):
    release_id, chart_set, charts, sandbox_name = get_chart_args(arguments)
    options = get_charts_options(arguments)
    return release_id, chart_set, charts, sandbox_name, options


def get_chart_args(arguments):
    release_id, chart_set, sandbox_name = get_release_args(arguments)
    charts = get_charts(arguments, chart_set)
    return release_id, chart_set, charts, sandbox_name


def get_charts_options(arguments):
    dry_run = arguments['--dry-run']
    debug = arguments['--debug']
    no_hooks = arguments['--no-hooks']
    return {'debug': debug, 'dry_run':dry_run, 'no_hooks': no_hooks}


def get_release_args(arguments):
    release_id = arguments['<release_id>']
    chart_set = arguments['<chart_set>']
    sandbox_name = arguments['--sandbox']
    return release_id, chart_set, sandbox_name


def get_charts(arguments, chart_set):
    charts = arguments['--charts']
    if charts is not None:
        charts = charts.split(',')
    else:
        charts = get_charts_to_use(chart_set)
    return charts


def get_get_current_cluster_args(arguments):
    output_type = arguments['--output']
    return output_type


def get_get_clusters_args(arguments):
    domain = arguments['<domain>']
    output_type = arguments['--output']
    if domain is None:
        domain = cluster.get_default_domain_from_cluster_config()
    return domain, output_type


def get_create_cluster_args(arguments):
    cluster_id = get_cluster_args(arguments)
    options = get_cluster_options(arguments)
    return cluster_id, options


def get_use_cluster_args(arguments):
    return get_cluster_args(arguments)


def get_validate_cluster_args(arguments):
    return get_cluster_args_default_to_current(arguments)


def get_delete_cluster_args(arguments):
    return get_cluster_args(arguments)


def get_cluster_args_default_to_current(arguments):
    cluster_id = get_cluster_args(arguments)
    if not cluster_id:
        cluster_id = cluster.get_current_cluster_id()
    return cluster_id


def get_cluster_args(arguments):
    cluster_id = arguments['<cluster_id>']
    return cluster_id


def get_cluster_options(arguments):
    debug = arguments['--debug']
    return {'debug': debug}


def get_use_namespace_args(arguments):
    namespace = arguments['<namespace>']
    return namespace



import time
from aloft.output import print_action
from aloft.process import execute
from subprocess import CalledProcessError


def wait_for_domain_to_resolve(domain):
    print_action(f'Waiting for domain to resolve: {domain}')
    success_count_in_a_row = 0

    while success_count_in_a_row < 10:
        if does_domain_resolve(domain):
            success_count_in_a_row += 1
            print('+', end='', flush=True)
        else:
            success_count_in_a_row = 0
            print('.', end='', flush=True)
        time.sleep(2)

    print(' resolved\n')


def does_domain_resolve(domain):
    resolves = True

    try:
        execute(f'host -t SOA {domain}', quiet=True)
    except CalledProcessError:
        resolves = False

    return resolves





import contextlib
import os
import time
import yaml
from aloft import aws
from aloft import chart
from aloft import cluster_config
from aloft import host
from aloft import chart_config
from aloft import output
from aloft.kubeconfig import kubeconfig
from aloft.process import execute
from aloft.vpc import delete_vpc, get_or_create_vpc
from subprocess import CalledProcessError


def get_and_print_current_cluster(output_type):
    current_cluster_id = get_current_cluster_id()
    cluster = get_cluster_by_cluster_id(current_cluster_id)
    print_clusters([cluster], output_type)


def get_current_cluster_id():
    stdout = execute('kubectl config current-context', quiet=True)
    return stdout.rstrip()


def get_and_print_clusters(cluster_domain, output_type):
    clusters = get_clusters(cluster_domain)
    print_clusters(clusters, output_type)


def print_clusters(clusters, output_type):
    if output_type == 'yaml':
        output.print_list_as_yaml(clusters)
    elif output_type == 'name':
        print_cluster_names(clusters)
    else:
        print_cluster_text(clusters)


def print_cluster_names(clusters):
    for cluster in clusters:
        print(cluster['metadata']['name'])


def print_cluster_text(clusters):
    rows = []

    for cluster in clusters:
        rows.append([
            cluster['metadata']['name'],
            cluster['spec']['kubernetesVersion'],
            cluster['spec']['networkCIDR'],
        ])

    output.print_table(['NAME', 'VERSION', 'CIDR'], rows, '{0: <40} {1: <10} {2: <18}')


def get_cluster_by_cluster_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    clusters = get_clusters_by_vpc_id(vpc_id)
    matching_clusters = list(filter(lambda cluster: cluster['metadata']['name'] == cluster_id, clusters))

    if len(matching_clusters) != 1:
        raise KeyError(f'Unable to get cluster data for: {cluster_id}')

    return matching_clusters[0]


def get_clusters(cluster_domain):
    vpc_names = get_vpc_names(cluster_domain)
    clusters = []

    for vpc_name in vpc_names:
        vpc_id = get_vpc_id_by_vpc_name_and_cluster_domain(vpc_name, cluster_domain)
        clusters += get_clusters_by_vpc_id(vpc_id)

    return clusters


def get_clusters_by_vpc_id(vpc_id):
    assume_role_by_vpc_id(vpc_id)
    state_store_locator = get_state_store_locator(vpc_id)
    stdout = execute(f'kops get clusters -o yaml --state {state_store_locator}', ['no clusters found'], quiet=True)
    clusters = []

    if stdout:
        for cluster_definition in stdout.split('---'):
            cluster = yaml.load(cluster_definition)
            clusters.append(cluster)

    return clusters


def use_cluster_for_release(release_id, chart_set, sandbox_name):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    cluster_id = release_config['cluster']
    use_cluster(cluster_id)


def use_cluster(cluster_id):
    assume_role_by_cluster_id(cluster_id)
    state_store_locator = get_state_store_locator_by_cluster_id(cluster_id)
    execute(f'kops export kubecfg {cluster_id} --state {state_store_locator}')


def delete_cluster(cluster_id):
    assume_role_by_cluster_id(cluster_id)
    state_store_locator = get_state_store_locator_by_cluster_id(cluster_id)
    execute(f'kops delete cluster {cluster_id} --yes --state {state_store_locator}')


def get_state_store_locator_by_cluster_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    return get_state_store_locator(vpc_id)


def get_state_store_locator(vpc_id):
    state_store_name = get_state_store_name(vpc_id)
    return f's3://{state_store_name}'


def get_state_store_name(vpc_id):
    state_store = f'k8s-state.{vpc_id}'
    bucket_name = get_state_store_bucket()
    return state_store if bucket_name is None else f'{bucket_name}/{state_store}'

def get_state_store_bucket():
    try:
     return os.environ['AWS_STATE_STORE_BUCKET']
    except KeyError:
        return None

def assume_role_by_cluster_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    assume_role_by_vpc_id(vpc_id)


def get_aws_profile_name_by_cluster_id(cluster_id):
    return get_vpc_name_by_custer_id(cluster_id)


def assume_role_by_vpc_id(vpc_id):
    aws_profile_name = get_aws_profile_name_by_vpc_id(vpc_id)
    aws.assume_role(aws_profile_name)


def get_aws_profile_name_by_vpc_id(vpc_id):
    return get_vpc_name_by_vpc_id(vpc_id)


def get_vpc_id_by_vpc_name_and_cluster_domain(vpc_name, cluster_domain):
    return f'{vpc_name}.{cluster_domain}'


def get_vpc_name_by_custer_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    return get_vpc_name_by_vpc_id(vpc_id)


def get_vpc_name_by_vpc_id(vpc_id):
    return vpc_id.split('.', 1)[0]


def get_vpc_id_by_cluster_id(cluster_id):
    return cluster_id.split('.', 1)[1]


def get_vpc_names(domain):
    domain_cluster_config_directory = get_domain_cluster_values_directory(domain)
    vpc_names = os.listdir(domain_cluster_config_directory)

    if len(vpc_names) == 0:
        raise ValueError(f'No cluster configuration found in {domain_cluster_config_directory}.')

    vpc_names.sort()

    return vpc_names


def get_domain(cluster_id):
    return cluster_id.split('.')[-2:]


def get_domain_cluster_values_directory(domain):
    base_cluster_config_directory = cluster_config.get_cluster_values_directory()
    return f'{base_cluster_config_directory}/{domain}'


def get_default_domain_from_cluster_config():
    base_cluster_config_directory = cluster_config.get_cluster_values_directory()
    domains = os.listdir(base_cluster_config_directory)

    if len(domains) > 1:
        raise ValueError(f'More than one domain config directory found in {base_cluster_config_directory}. You must '
                         f'specify a <domain>')
    elif not domains:
        raise ValueError(f'No cluster domain configurations found in {base_cluster_config_directory}.')

    return domains[0]


def conditionally_create_vpc_and_cluster(cluster_id, options):
    vpc = get_or_create_vpc_by_cluster_id(cluster_id)
    create_cluster_and_attach_network(cluster_id, vpc, options)


def create_cluster_and_attach_network(cluster_id, vpc, options):
    vpn_name = get_vpc_name_by_custer_id(cluster_id)
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    state_store_name = get_state_store_name(vpc_id)

    aws.conditionally_create_hosted_zone_and_setup_ns_records(cluster_id, vpc_id)
    host.wait_for_domain_to_resolve(cluster_id)
    aws.get_or_create_s3_bucket(state_store_name)
    create_cluster(cluster_id, vpc, options)
    propagate_cluster_route_tables(cluster_id, vpn_name)
    add_cluster_node_policies(cluster_id)
    wait_for_cluster_to_be_ready(cluster_id)
    initialize_helm()


def propagate_cluster_route_tables(cluster_id, vpn_name):
    vpn_gateway_id = aws.get_vpn_gateway_id(vpn_name)
    route_table_ids = aws.get_route_table_ids(cluster_id)

    for route_table_id in route_table_ids:
        aws.propagate_route_table(route_table_id, vpn_gateway_id)


def add_cluster_node_policies(cluster_id):
    node_policies_filename = cluster_config.get_node_policies_filename()

    with open(node_policies_filename, 'r') as stream:
        node_policies = stream.read()
        aws.set_role_policy(f'nodes.{cluster_id}', 'systemPolicy', node_policies)


def wait_for_cluster_to_be_ready(cluster_id):
    output.print_action(f'Waiting for {cluster_id} to be ready.')
    time.sleep(300)

    while not validate_cluster(cluster_id):
        time.sleep(20)


def validate_cluster(cluster_id):
    state_store_locator = get_state_store_locator_by_cluster_id(cluster_id)
    valid = False

    try:
        execute(f'kops validate cluster {cluster_id} --state {state_store_locator}')
        valid = True
    except CalledProcessError:
        pass

    return valid


def create_cluster_and_apply_system(cluster_id, options):
    home = os.environ['HOME']
    with kubeconfig(f'{home}/.kube/config.{cluster_id}') as config:
        conditionally_create_vpc_and_cluster(cluster_id, options)
        chart.apply_charts(cluster_id, 'system', ['system'], None, options)
    use_cluster(cluster_id)


def create_cluster(cluster_id, vpc, options):
    output.print_action(f'Creating cluster {cluster_id} in {vpc["Name"]} vpc.')
    state_store_locator = get_state_store_locator_by_cluster_id(cluster_id)
    generated_cluster_config_filename = cluster_config.generated_cluster_config_filename(cluster_id)
    generate_cluster_config_from_template(cluster_id, vpc, generated_cluster_config_filename)

    execute(f'kops create -f {generated_cluster_config_filename} --state {state_store_locator}')
    create_ssh_secret(cluster_id, state_store_locator, options)
    execute(f'kops update cluster {cluster_id} --yes --state {state_store_locator}')


def generate_cluster_config_from_template(cluster_id, vpc, generated_cluster_config_filename):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    cluster_template_filename = cluster_config.get_cluster_template_filename()
    default_cluster_values_filename = cluster_config.get_default_cluster_values_filename()
    cluster_values_filename = cluster_config.get_cluster_values_filename(cluster_id)
    vpc_values_filename = cluster_config.get_vpc_config_filename(vpc_id)
    dynamic_values_filename = cluster_config.get_create_dynamic_values_filename(cluster_id)

    create_dynamic_values_file(dynamic_values_filename, vpc, cluster_id)

    execute(f'kops toolbox template '
            f'--values {default_cluster_values_filename} '
            f'--values {cluster_values_filename} '
            f'--values {vpc_values_filename} '
            f'--values {dynamic_values_filename} '
            f'--template {cluster_template_filename} '
            f'--format-yaml=true > {generated_cluster_config_filename}')


def create_dynamic_values_file(vpc_id_value_filename, vpc, cluster_id):
    output.write_to_file(
        vpc_id_value_filename,
        f'networkId: {vpc["VpcId"]}\n'
        f'clusterFQDN: {cluster_id}\n'
        f'kopsStateStore: {get_state_store_locator_by_cluster_id(cluster_id)}\n'
    )


def create_ssh_secret(cluster_id, state_store_locator, options):
    key_filename = '/tmp/cluster_rsa'
    create_cluster_ssh_keys(key_filename)
    execute(f'kops create secret --name {cluster_id} sshpublickey admin'
            f' -i "{key_filename}.pub" --state {state_store_locator}')
    if not options.get('debug'):
        remove_cluster_ssh_keys(key_filename)


def initialize_helm():
    execute('helm init')
    execute('kubectl rollout status -w deployment/tiller-deploy --namespace=kube-system')


def create_cluster_ssh_keys(key_filename):
    remove_cluster_ssh_keys(key_filename)
    execute(f'ssh-keygen -q -f {key_filename} -t rsa -N ""')


def remove_cluster_ssh_keys(key_filename):
    remove(key_filename)
    remove(f'{key_filename}.pub')


def remove(filename):
    with contextlib.suppress(FileNotFoundError):
        os.remove(filename)


def get_or_create_vpc_by_cluster_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    return get_or_create_vpc(vpc_id)


def delete_vpc_if_no_clusters_exist(vpc_id):
    clusters = get_clusters_by_vpc_id(vpc_id)

    if len(clusters) > 0:
        output.print_action(f'Not deleting VPC. All clusters must be deleted before deleting the VPC.')
        print_cluster_names(clusters)
        print()
    else:
        delete_vpc(vpc_id)



import os
import base64

from aloft.process import execute
from aloft import aws
from aloft import k8s
from aloft import chart_config
from aloft import output
from aloft import volume
from tempfile import NamedTemporaryFile


def apply_charts(release_id, chart_set, charts, sandbox_name, options):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    namespace = release_config['namespace']
    if not options.get('dry_run'):
        k8s.create_namespace(namespace)
        volume.restore_volumes(release_id, chart_set, charts, sandbox_name)

    for chart in charts:
        value_filenames = chart_config.generate_value_files(release_id, chart_set, chart, release_config)
        config = chart_config.get_chart_config_for_chart(chart_set, chart)
        secrets_config = chart_config.get_secrets_config_for_chart(chart_set, chart)
        apply_chart(chart, config, namespace, value_filenames, secrets_config, options)

    if not options.get('dry_run') and release_config.get('lockVolumes', False):
        volume.lock_volumes(release_id, chart_set, charts, sandbox_name)


def apply_chart(chart, config, namespace, value_filenames, secrets_config, options):
    base_chart_config_directory = chart_config.get_base_chart_config_directory()
    chart_directory = f'{base_chart_config_directory}/charts/{chart}'
    value_args = ' '.join(map(lambda filename: f'-f {filename}', value_filenames))
    release_name = get_release_name(chart, namespace)
    options['helm_timout'] = config.get('helmTimeout')
    helm_args = get_helm_args(options)

    if not options.get('dry_run'):
        create_secrets(chart, secrets_config, namespace)
        if not options.get('no_hooks'):
            execute_install_hook('pre-install', chart_directory)
    execute(f'helm dependencies build {chart_directory}')
    execute(f'helm upgrade {helm_args}-i {release_name} --namespace {namespace} {chart_directory} {value_args}')
    if not options.get('dry_run') and not options.get('no_hooks'):
        execute_install_hook('post-install', chart_directory)


def delete_charts(release_id, chart_set, charts, sandbox_name, options):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    namespace = release_config['namespace']

    for chart in charts:
        secrets_config = chart_config.get_secrets_config_for_chart(chart_set, chart)
        delete_chart(chart, namespace, secrets_config, options)

    if not options.get('dry_run'):
        k8s.delete_namespace_if_empty(namespace)
        volume.remove_released_volume_resources(release_id, chart_set, charts, sandbox_name)


def delete_chart(chart, namespace, secrets_config, options):
    base_chart_config_directory = chart_config.get_base_chart_config_directory()
    chart_directory = f'{base_chart_config_directory}/charts/{chart}'
    release_name = get_release_name(chart, namespace)
    helm_args = get_helm_args(options)

    if not options.get('dry_run') and not options.get('no_hooks'):
        execute_install_hook('pre-uninstall', chart_directory)
    execute(f'helm delete {helm_args}--purge {release_name}', ['not found'])
    if not options.get('dry_run') and not options.get('no_hooks'):
        delete_secrets(chart, secrets_config, namespace)
        execute_install_hook('post-uninstall', chart_directory)


def get_helm_args(options):
    helm_args = ''

    if options:
        helm_timeout = options.get('helm_timout')
        if helm_timeout:
            helm_args += f'--timeout {helm_timeout} '
        if options.get('debug'):
            helm_args += '--debug '
        if options.get('dry_run'):
            helm_args += '--dry-run '
        if options.get('no_hooks'):
            helm_args += '--no-hooks '

    return helm_args


def get_release_name(chart, namespace):
    release_name = chart

    if chart != namespace:
        release_name = f'{namespace}-{chart}'

    return release_name


def create_secrets(chart, secrets_config, namespace):

    if secrets_config:
        create_standard_secrets(secrets_config, chart, namespace)
        create_docker_secrets(secrets_config, namespace)


def create_standard_secrets(secrets_config, chart, namespace):
    secret_args = ''
    secret_keys = []
    temp_filenames = []

    try:
        for item in secrets_config.get('items', []):
            vault_key = item.get('vaultKey', None)
            secret_key = item.get('secretKey', None)
            encoded = item.get('encoded', False)

            if secret_key:
                secret_keys.append(secret_key)
                secret_value = aws.get_secret(vault_key)

                if secret_value:
                    if encoded:
                        secret_value = base64.standard_b64decode(secret_value).decode('utf-8')

                    temp_file = NamedTemporaryFile(mode='w+t', delete=False)
                    temp_file.write(secret_value)
                    temp_file.flush()
                    temp_filenames.append(temp_file.name)
                    secret_args = f'{secret_args} --from-file={secret_key}={temp_file.name}'

        if secret_args:
            secrets_name = secrets_config.get('secretName', f'{chart}-secret')
            output.print_action(f'Creating secret {secrets_name} with keys {secret_keys}')
            execute(f'kubectl -n {namespace} delete secret {secrets_name}', ['NotFound'], quiet=False)
            execute(f'kubectl -n {namespace} create secret generic {secrets_name} {secret_args}', quiet=False)

    finally:
        if temp_filenames:
            rm_args = ' '.join(temp_filenames)
            execute(f'rm -f {rm_args}')


def create_docker_secrets(secrets_config, namespace):
    for secret in secrets_config.get('dockerRegistrySecrets', []):
        vault_username_key = secret.get('vaultUsernameKey', None)
        vault_password_key = secret.get('vaultPasswordKey', None)

        docker_username = aws.get_secret(vault_username_key)
        docker_password = aws.get_secret(vault_password_key)

        output.print_action(f'Creating docker-registry secret {secret["secretName"]}'
                            f' with username {vault_username_key} and password {vault_password_key}')
        execute(f'kubectl -n {namespace} delete secret {secret["secretName"]}', ['NotFound'], quiet=True)
        execute(f'kubectl -n {namespace} create secret docker-registry {secret["secretName"]}'
                f' --docker-server={secret["dockerServer"]} --docker-username={docker_username}'
                f' --docker-password={docker_password} --docker-email={secret["dockerEmail"]}', quiet=True)


def delete_secrets(chart, secrets_config, namespace):
    if secrets_config:
        secrets_name = secrets_config.get('secretName', f'{chart}-secret')
        execute(f'kubectl -n {namespace} delete secret {secrets_name}', ['NotFound'])

        for secret in secrets_config.get('dockerRegistrySecrets', []):
            execute(f'kubectl -n {namespace} delete secret {secret["secretName"]}', ['NotFound'])


def execute_install_hook(hook_type, chart_directory):
    hook_script = f'{chart_directory}/hooks/{hook_type}'

    if os.path.exists(hook_script):
        execute(hook_script)



import yaml
from aloft.process import execute


def apply_file(filename):
    execute(f'kubectl apply -f {filename}')


def create_namespace(namespace):
    execute(f'kubectl create namespace {namespace}', ['AlreadyExists'])


def delete_namespace_if_empty(namespace):
    releases = get_releases_by_namespace(namespace)

    if not releases:
        delete_namespace(namespace)


def delete_namespace(namespace):
    execute(f'kubectl delete namespace {namespace}', ['NotFound'])


def use_namespace(namespace):
    execute(f'kubectl config set-context $(kubectl config current-context) --namespace={namespace}')


def get_releases_by_namespace(namespace):
    stdout = execute(f'helm ls -q --namespace {namespace}')
    return stdout.splitlines()


def get_resource_by_name(resource_type, resource_name, namespace=None):
    resource = None
    resources = get_resources(resource_type, namespace, None, resource_name)

    if len(resources) == 1:
        resource = resources[0]

    return resource


def get_resources(resource_type, namespace=None, label_selectors=None, resource_name=None):
    label_arg = ''
    namespace_arg = ''
    resource_name_arg = ''
    resources = []

    if resource_name:
        resource_name_arg = f'{resource_name} '

    if namespace:
        namespace_arg = f'-n {namespace} '

    if label_selectors:
        label_format = '{key}{operator}{value}'
        label_parameters = ",".join(map(lambda selector: label_format.format_map(selector), label_selectors))
        label_arg = f'-l {label_parameters} '

    stdout = execute(f'kubectl get {resource_type} {resource_name_arg}{namespace_arg}{label_arg}-o yaml', ['NotFound'])

    if stdout:
        resource_data = yaml.load(stdout)
        resources = resource_data.get('items', [resource_data])

    return resources



import time
from aloft.output import print_action, print_details, print_details_line
from subprocess import Popen, PIPE, CalledProcessError


def execute(cmd, ignore_error_messages=None, quiet=False, retry_error_messages=['x509']):
    process = Popen(cmd, stderr=PIPE, stdout=PIPE, shell=True, encoding='utf-8')
    stderr, stdout = communicate(process, cmd, quiet)
    return_code = process.returncode
    should_ignore_error = should_ignore_return_code(return_code, ignore_error_messages, stderr, quiet)
    print_stderr(stderr, return_code, should_ignore_error)

    if should_retry(return_code, retry_error_messages, stderr, quiet):
        time.sleep(2)
        return execute(cmd, ignore_error_messages, quiet, retry_error_messages)
    else:
        if not should_ignore_error:
            raise_if_error(return_code, cmd, stdout, stderr)

    return stdout


def communicate(process, cmd, quiet):
    stdout = ""

    if not quiet:
        print_action(cmd)

    for line in process.stdout:
        if not quiet:
            print_details_line(line.rstrip())
        stdout += line

    if not quiet and stdout:
        print('')

    stdout_empty, stderr = process.communicate()

    return stderr, stdout


def print_stderr(stderr, return_code, ignoring):
    if stderr and return_code != 0 and not ignoring:
        print("\nERROR:")
        print_details(stderr)


def should_ignore_return_code(return_code, ignore_error_messages, stderr, quiet):
    ignore = False
    if return_code != 0 and ignore_error_messages:
        for ignore_error_message in ignore_error_messages:
            if ignore_error_message in stderr:
                if not quiet:
                    print_details(f'Ignoring return code: {return_code} - {ignore_error_message}')
                ignore = True
                break

    return ignore


def should_retry(return_code, retry_error_messages, stderr, quiet):
    retry = False

    if return_code != 0 and retry_error_messages:
        for retry_error_message in retry_error_messages:
            if retry_error_message in stderr:
                if not quiet:
                    print_details(f'Retrying: {return_code} - {retry_error_message}')
                    retry = True
                break

    return retry

def raise_if_error(return_code, cmd, stdout, stderr):
    if return_code != 0:
        raise CalledProcessError(return_code, cmd, stdout, stderr)



from aloft import input


def get_profile():
    return get_aloft_config()['profile']


def get_roles():
    return get_aloft_config()['roles']


def get_aloft_config():
    return input.read_yaml_file(get_aloft_config_filename())


def get_aloft_config_filename():
    aloft_config_directory = get_aloft_config_directory()
    return f'{aloft_config_directory}/config.yaml'


def get_aloft_config_directory():
    return input.get_directory_from_environment('ALOFT_CONFIG', '~/.aloft')



import boto3
import os
import re
import time
import uuid
from aloft.output import print_action
from aloft import config
from botocore.exceptions import ClientError

credentials = {}
region = None


def get_vpc_by_name(vpc_name):
    client = get_ec2_client()
    vpc = None

    vpcs = client.describe_vpcs(
        Filters=[
            {'Name': 'tag-key', 'Values': ['Name']},
            {'Name': 'tag-value', 'Values': [vpc_name]}
        ]
    )

    if len(vpcs['Vpcs']) == 1:
        vpc = vpcs['Vpcs'][0]
        vpc['Name'] = vpc_name

    return vpc


def assume_role(role_name):
    global credentials
    roles = config.get_roles()
    role = roles.get(role_name, None)

    os.environ['AWS_PROFILE'] = config.get_profile()
    if role and role.get('arn', None):
        client = get_sts_client()
        try:
            assumed_role = client.assume_role(
                RoleArn=role['arn'],
                RoleSessionName=role_name
            )
            set_credentials(assumed_role['Credentials'])
            set_region(role.get('region', 'us-east-1'))
        except ClientError as error:
            if error.response['Error']['Code'] == 'ExpiredToken':
                print(error)
                exit(2)
    else:
        raise KeyError(f'Unable to find role {role_name}')


def set_credentials(new_credentials):
    global credentials
    credentials = new_credentials
    os.environ['AWS_ACCESS_KEY_ID'] = credentials['AccessKeyId']
    os.environ['AWS_SECRET_ACCESS_KEY'] = credentials['SecretAccessKey']
    os.environ['AWS_SESSION_TOKEN'] = credentials['SessionToken']


def set_region(new_region):
    global region
    region = new_region
    os.environ['AWS_REGiON'] = region


def delete_ec2_tags(resource_id, keys_to_delete):
    client = get_ec2_client()
    tags = []

    for key in keys_to_delete:
        tags.append({'Key': key})

    client.delete_tags(Resources=[resource_id], Tags=tags)


def create_ec2_tags(resource_id, tags_to_create_dictionary):
    client = get_ec2_client()
    tags = []

    for key, value in tags_to_create_dictionary.items():
        tags.append({'Key': key, 'Value': value})

    client.create_tags(Resources=[resource_id], Tags=tags)


def save_content_to_s3(content, bucket_name, key):
    bucket = get_or_create_s3_bucket(bucket_name)
    bucket.put_object(Body=content, Key=key)


def get_content_from_s3(bucket_name, key):
    s3 = get_s3_resource()
    s3_object = s3.Object(bucket_name, key)
    content = None

    try:
        content = s3_object.get()['Body'].read().decode('utf-8')
    except s3.meta.client.exceptions.NoSuchKey:
        pass
    except s3.meta.client.exceptions.NoSuchBucket:
        pass

    return content


def delete_file_in_s3(bucket_name, key):
    bucket = get_s3_bucket(bucket_name)
    objects_to_delete = {'Objects': [{'Key': key}]}
    bucket.delete_objects(Delete=objects_to_delete)


def get_or_create_s3_bucket(bucket_name):
    bucket = get_s3_bucket(bucket_name)

    if not bucket:
        create_s3_bucket(bucket_name)
        bucket = get_s3_bucket(bucket_name)

    return bucket


def get_s3_bucket(bucket_name):
    buckets = get_all_buckets()
    filtered_buckets = list(filter(lambda x: x.name == bucket_name, buckets))
    bucket = None

    if len(filtered_buckets) == 1:
        bucket = filtered_buckets[0]

    return bucket


def create_s3_bucket(bucket_name):
    s3 = get_s3_resource()
    session = boto3.session.Session()
    region_name = session.region_name

    if region_name == 'us-east-1':
        s3.create_bucket(Bucket=bucket_name)
    else:
        s3.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={'LocationConstraint': region_name})


def get_all_buckets():
    s3 = get_s3_resource()
    return s3.buckets.all()


def get_vpn_gateway_id(vpn_name):
    vpn_gateway_id = None
    vpn_gateway = get_vpn_gateway(vpn_name)

    if vpn_gateway:
        vpn_gateway_id = vpn_gateway['VpnGatewayId']

    return vpn_gateway_id


def get_vpn_gateway_attachment_status(vpc):
    aws_vpc_id = vpc['VpcId']
    vpn_gateway = get_vpn_gateway(vpc['Name'])
    status = None

    if vpn_gateway:
        vpn_attachments = vpn_gateway.get('VpcAttachments', None)

        filtered_vpn_attachments = list(filter(lambda attachment: attachment['VpcId'] == aws_vpc_id, vpn_attachments))

        if filtered_vpn_attachments and len(filtered_vpn_attachments) == 1:
            status = filtered_vpn_attachments[0]['State']

    return status


def get_vpn_gateway(vpn_name):
    client = get_ec2_client()
    vpn_gateway = None

    vpn_gateways = client.describe_vpn_gateways(
        Filters=[
            {'Name': 'tag-key', 'Values': ['Name']},
            {'Name': 'tag-value', 'Values': [f'{vpn_name}-network']}
        ]
    )

    if len(vpn_gateways['VpnGateways']) == 1:
        vpn_gateway = vpn_gateways['VpnGateways'][0]

    return vpn_gateway


def get_route_table_ids(cluster_id):
    client = get_ec2_client()
    route_table_ids = None

    route_tables = client.describe_route_tables(
        Filters=[
            {'Name': 'tag-key', 'Values': ['KubernetesCluster']},
            {'Name': 'tag-value', 'Values': [cluster_id]}
        ]
    )

    if len(route_tables['RouteTables']) > 0:
        route_table_ids = list(map(lambda route_table: route_table['RouteTableId'], route_tables['RouteTables']))

    return route_table_ids


def propagate_route_table(route_table_id, vpn_gateway_id):
    print_action(f'Propagating route table: {route_table_id} {vpn_gateway_id}')

    client = get_ec2_client()
    client.enable_vgw_route_propagation(
        GatewayId=vpn_gateway_id,
        RouteTableId=route_table_id
    )


def set_role_policy(role_name, policy_name, policy_document):
    print_action(f'Setting role policy: {role_name} {policy_name}')

    client = get_iam_client()
    client.put_role_policy(
        RoleName=role_name,
        PolicyName=policy_name,
        PolicyDocument=policy_document
    )


def conditionally_create_hosted_zone_and_setup_ns_records(hosted_zone_to_create, parent_hosted_zone):
    hosted_zone = get_hosted_zone(hosted_zone_to_create)

    if not hosted_zone:
        print_action(f'Creating hosted zone: {hosted_zone_to_create}')
        new_hosted_zone = create_hosted_zone(hosted_zone_to_create)
        setup_ns_records_in_parent_domain(hosted_zone_to_create, parent_hosted_zone, new_hosted_zone)
    else:
        print_action(f'Hosted zone for {hosted_zone_to_create} already exists.  Not recreating hosted DNS zone.')


def setup_ns_records_in_parent_domain(cluster_id, vpc_id, child_hosted_zone):
    vpc_hosted_zone = get_hosted_zone(vpc_id)
    add_ns_records_to_hosted_zone(vpc_hosted_zone['Id'], cluster_id, child_hosted_zone['DelegationSet']['NameServers'])


def get_hosted_zone(domain):
    client = get_route53_client()
    hosted_zone = None

    hosted_zones = client.list_hosted_zones_by_name(
        DNSName=domain,
        MaxItems='1'
    )

    if (len(hosted_zones['HostedZones']) == 1 and
            re.match(f'{domain}[.]?', hosted_zones['HostedZones'][0]['Name']) is not None):
        hosted_zone = hosted_zones['HostedZones'][0]

    return hosted_zone


def create_hosted_zone(domain):
    client = get_route53_client()
    return client.create_hosted_zone(Name=domain, CallerReference=f'{uuid.uuid4()}')


def add_ns_records_to_hosted_zone(hosted_zone_id, record_set_name, ns_records):
    client = get_route53_client()
    resource_records = list(map(lambda record: {'Value': record}, ns_records))

    client.change_resource_record_sets(
        HostedZoneId=hosted_zone_id,
        ChangeBatch={
            'Changes': [
                {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'Name': record_set_name,
                        'Type': 'NS',
                        'TTL': 300,
                        'ResourceRecords': resource_records
                    }
                },
            ]
        }
    )


def create_vpc(vpc_name, cidr_block):
    print_action(f'Creating VPC {vpc_name} using CIDR {cidr_block}')
    client = get_ec2_client()
    response = client.create_vpc(
        CidrBlock=cidr_block,
    )

    vpc = response.get('Vpc', None)

    if vpc:
        vpc['Name'] = vpc_name
        create_ec2_tags(vpc['VpcId'], {'Name': vpc_name})

    return vpc


def create_and_attach_internet_gateway(vpc):
    internet_gateway = create_internet_gateway(vpc['Name'])
    attach_internet_gateway_to_vpc(internet_gateway, vpc)


def create_internet_gateway(vpc_name):
    print_action(f'Creating internet gateway for {vpc_name}')
    client = get_ec2_client()
    response = client.create_internet_gateway()

    internet_gateway = response.get('InternetGateway', None)

    if internet_gateway:
        create_ec2_tags(internet_gateway['InternetGatewayId'], {'Name': vpc_name})

    return internet_gateway


def attach_internet_gateway_to_vpc(internet_gateway, vpc):
    print_action('Attaching internet gateway to vpc.')
    client = get_ec2_client()
    client.attach_internet_gateway(
        InternetGatewayId=internet_gateway['InternetGatewayId'],
        VpcId=vpc['VpcId']
    )


def create_vpn_gateway(vpc_name):
    client = get_ec2_client()
    response = client.create_vpn_gateway(
        Type='ipsec.1',
    )

    vpn_gateway = response.get('VpnGateway', None)

    if vpn_gateway:
        create_ec2_tags(
            vpn_gateway['VpnGatewayId'],
            {
                'Name': f'{vpc_name}-network',
                'transitVPC1:preferred-path': 'CSR1',
                'transitVPC1:spoke': 'true',
            }
        )

    return vpn_gateway


def create_and_attach_vpn_gateway(vpc):
    vpn_gateway = get_vpn_gateway(vpc['Name'])

    if not vpn_gateway:
        print_action('Creating vpn gateway.')
        vpn_gateway = create_vpn_gateway(vpc['Name'])

    attach_vpn_gateway_and_wait_until_ready(vpn_gateway, vpc)


def attach_vpn_gateway_and_wait_until_ready(vpn_gateway, vpc):
    attach_vpn_gateway(vpn_gateway, vpc)
    print_action('Waiting for vpn gateway to attach to vpc.')
    time.sleep(20)

    while get_vpn_gateway_attachment_status(vpc) != 'attached':
        print('.', end='', flush=True)
        time.sleep(10)

    print(' attached\n')


def attach_vpn_gateway(vpn_gateway, vpc):
    client = get_ec2_client()
    response = None
    state = None

    try:
        response = client.attach_vpn_gateway(
            VpnGatewayId=vpn_gateway['VpnGatewayId'],
            VpcId=vpc['VpcId']
        )
    except ClientError as error:
        if error.response['Error']['Code'] != 'InvalidVpcID.NotFound':
            print(error)
            raise error

    if response:
        attachment = response.get('VpcAttachment', {'State': None})
        state = attachment['State']

    return state


def delete_network_and_vpc(vpc_name):
    vpc = get_vpc_by_name(vpc_name)
    vpn_gateway = get_vpn_gateway(vpc_name)
    vpn_connections = get_vpn_connections(vpn_gateway)
    internet_gateway = get_internet_gateway(vpc_name)

    remove_vpn_gateway_tags(vpn_gateway)
    detach_vpn_gateway_and_wait_until_detached(vpn_gateway, vpc)
    delete_vpn_connections(vpn_connections)
    delete_vpn_gateway(vpn_gateway)
    detach_internet_gateway(internet_gateway, vpc)
    delete_internet_gateway(internet_gateway)
    delete_vpc(vpc)


def get_vpn_connections(vpn_gateway):
    vpn_connections = None

    if vpn_gateway:
        client = get_ec2_client()
        response = client.describe_vpn_connections(
            Filters=[
                {
                    'Name': 'vpn-gateway-id',
                    'Values': [
                        vpn_gateway['VpnGatewayId'],
                    ]
                },
            ]
        )

        vpn_connections = response.get('VpnConnections', None)

    return vpn_connections


def delete_vpn_connections(vpn_connections):
    if vpn_connections:
        client = get_ec2_client()
        for vpn_connection in vpn_connections:
            print_action(f'Deleting VPN connection: {vpn_connection["VpnConnectionId"]}')
            client.delete_vpn_connection(
                VpnConnectionId=vpn_connection['VpnConnectionId']
            )


def delete_vpn_gateway(vpn_gateway):
    if vpn_gateway:
        print_action(f'Deleting VPN gateway: vpn_gateway["VpnGatewayId"]')
        client = get_ec2_client()
        client.delete_vpn_gateway(
            VpnGatewayId=vpn_gateway['VpnGatewayId']
        )


def get_internet_gateway(name):
    client = get_ec2_client()
    internet_gateway = None
    response = client.describe_internet_gateways(
        Filters=[
            {'Name': 'tag-key', 'Values': ['Name']},
            {'Name': 'tag-value', 'Values': [name]}
        ]
    )

    internet_gateways = response.get('InternetGateways', None)

    if internet_gateways and len(internet_gateways) == 1:
        internet_gateway = internet_gateways[0]

    return internet_gateway


def detach_internet_gateway(internet_gateway, vpc):
    if vpc and internet_gateway:
        print_action(f'Detaching internet gateway: {internet_gateway["InternetGatewayId"]}')
        client = get_ec2_client()
        client.detach_internet_gateway(
            VpcId=vpc['VpcId'],
            InternetGatewayId=internet_gateway['InternetGatewayId']
        )


def delete_internet_gateway(internet_gateway):
    if internet_gateway:
        print_action(f'Deleting internet gateway: {internet_gateway["InternetGatewayId"]}')
        client = get_ec2_client()
        client.delete_internet_gateway(
            InternetGatewayId=internet_gateway['InternetGatewayId']
        )


def delete_vpc(vpc):
    if vpc:
        client = get_ec2_client()
        deleted = False
        retries = 0
        print_action(f'Deleting VPC: {vpc["VpcId"]}')

        while not deleted or retries > 10:
            try:
                client.delete_vpc(
                    VpcId=vpc['VpcId'],
                )
                deleted = True
            except ClientError as error:
                if error.response['Error']['Code'] != 'DependencyViolation':
                    raise error
            retries += 1
            time.sleep(10)


def remove_vpn_gateway_tags(vpn_gateway):
    if vpn_gateway:
        print_action(f'Removing vpn gateway tags from: {vpn_gateway["VpnGatewayId"]}')
        delete_ec2_tags(vpn_gateway['VpnGatewayId'], ['transitVPC1:preferred-path', 'transitVPC1:spoke'])


def detach_vpn_gateway_and_wait_until_detached(vpn_gateway, vpc):
    if vpn_gateway:
        print_action('Waiting for vpn gateway to detach from vpc.')
        detach_vpn_gateway(vpn_gateway, vpc)
        time.sleep(20)

        while get_vpn_gateway_attachment_status(vpc) != 'detached':
            print('.', end='', flush=True)
            time.sleep(20)

        print(' detached\n')


def detach_vpn_gateway(vpn_gateway, vpc):
    client = get_ec2_client()
    try:
        client.detach_vpn_gateway(
            VpcId=vpc['VpcId'],
            VpnGatewayId=vpn_gateway['VpnGatewayId']
        )
    except ClientError as error:
        if error.response['Error']['Code'] != 'InvalidVpnGatewayAttachment.NotFound':
            raise error


def get_vpn_gateway_attach_status(vpn_gateway, vpc):
    client = get_ec2_client()
    response = client.detach_vpn_gateway(
        VpcId=vpc['VpcId'],
        VpnGatewayId=vpn_gateway['VpnGatewayId']
    )

    attachment = response.get('VpcAttachment', None)
    return attachment['State']


def get_secret(secret_name):
    client = get_ssm_client()
    secret = None

    response = client.get_parameter(
        Name=secret_name,
        WithDecryption=True
    )

    parameter = response.get('Parameter', None)

    if parameter:
        secret = parameter.get('Value', None)

    return secret


def get_ec2_client():
    return get_client('ec2')


def get_iam_client():
    return get_client('iam')


def get_s3_resource():
    return get_resource('s3')


def get_route53_client():
    return get_client('route53')


def get_ssm_client():
    return get_client('ssm')


def get_sts_client():
    return boto3.client('sts')


def get_client(client_type):
    return boto3.client(
        client_type,
        aws_access_key_id=credentials['AccessKeyId'],
        aws_secret_access_key=credentials['SecretAccessKey'],
        aws_session_token=credentials['SessionToken'],
        region_name=region
    )


def get_resource(resource_type):
    return boto3.resource(
        resource_type,
        aws_access_key_id=credentials['AccessKeyId'],
        aws_secret_access_key=credentials['SecretAccessKey'],
        aws_session_token=credentials['SessionToken'],
        region_name=region
    )


import jinja2
import os
import yaml


def get_directory_from_environment(environment_key, default_directory_name=None):
    directory_name = os.path.expanduser(os.environ.get(environment_key, default_directory_name))
    validate_config_directory(directory_name, environment_key)
    return directory_name


def validate_config_directory(config_directory, environment_key):
    if config_directory is None:
        raise NotADirectoryError(f'{environment_key} environment variable has not been set')
    elif not os.path.isdir(config_directory):
        raise NotADirectoryError(f'{environment_key} must reference a directory: {config_directory}')


def render_template_file(template_filename, context):
    path, filename = os.path.split(template_filename)
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(path)
    ).get_template(filename).render(context)


def read_yaml_file(filename):
    contents = None

    if os.path.isfile(filename):
        with open(filename, 'r') as stream:
            contents = yaml.load(stream)

    return contents



import yaml
from io import open
from aloft import config


def get_cluster_config_values(cluster_id):
    cluster_config_filename = get_cluster_values_filename(cluster_id)
    with open(cluster_config_filename, 'r') as stream:
        return yaml.load(stream)


def get_cluster_values_filename(cluster_id):
    cluster_config_directory = get_cluster_values_directory()
    cluster_name, vpc_name, domain_name = split_cluster_id(cluster_id)
    return f'{cluster_config_directory}/{domain_name}/{vpc_name}/{cluster_name}/cluster.yaml'


def get_default_cluster_values_filename():
    cluster_template_directory = get_cluster_template_directory()
    return f'{cluster_template_directory}/values.yaml'


def get_vpc_config_filename(vpc_id):
    cluster_config_directory = get_cluster_values_directory()
    vpc_name, domain_name = split_vpc_id(vpc_id)
    return f'{cluster_config_directory}/{domain_name}/{vpc_name}/vpc.yaml'


def get_create_dynamic_values_filename(cluster_id):
    return f'/tmp/{cluster_id}-dynamic.yaml'


def generated_cluster_config_filename(cluster_id):
    return f'/tmp/{cluster_id}.yaml'


def get_cluster_template_filename():
    cluster_template_directory = get_cluster_template_directory()
    return f'{cluster_template_directory}/template.yaml'


def get_node_policies_filename():
    cluster_template_directory = get_cluster_template_directory()
    return f'{cluster_template_directory}/node-policies.json'


def split_cluster_id(cluster_id):
    parts = cluster_id.split('.')
    cluster_name = parts[0]
    vpc_name = parts[1]
    domain_name = '.'.join(parts[2:4])
    return cluster_name, vpc_name, domain_name


def split_vpc_id(vpc_id):
    parts = vpc_id.split('.')
    vpc_name = parts[0]
    domain_name = '.'.join(parts[1:3])
    return vpc_name, domain_name


def get_cluster_template_directory():
    base_chart_config_directory = get_base_cluster_config_directory()
    return f'{base_chart_config_directory}/template'


def get_cluster_values_directory():
    base_chart_config_directory = get_base_cluster_config_directory()
    return f'{base_chart_config_directory}/values'


def get_base_cluster_config_directory():
    aloft_config_directory = config.get_aloft_config_directory()
    return f'{aloft_config_directory}/clusters'



import os
from aloft.output import print_action


class kubeconfig(object):

    def __init__(self, kubeconfig_filename):
        print_action(f'Using kubeconfig {kubeconfig_filename}')
        self.temporary_kubeconfig_filename = kubeconfig_filename

    def __enter__(self):
        self.original_kubeconfig = self.set_kubeconfig(self.temporary_kubeconfig_filename)
        return self.temporary_kubeconfig_filename

    def __exit__(self, type, value, traceback):
        self.set_kubeconfig(self.original_kubeconfig)

    @staticmethod
    def set_kubeconfig(kubeconfig_filename):
        original_kubeconfig = os.environ.get('KUBECONFIG', None)

        if kubeconfig_filename:
            os.environ['KUBECONFIG'] = kubeconfig_filename
        else:
            del os.environ['KUBECONFIG']

        return original_kubeconfig



import yaml


def print_action(action):
    print(f'**** {action}\n')


def print_details(details):
    for line in iter(details.splitlines()):
        print_details_line(line)
    print('')


def print_details_line(line):
    print(f'     {line.rstrip()}')


def print_table(headings, rows, row_format):
    print(row_format.format(*headings))

    for row in rows:
        print(row_format.format(*row))


def print_list_as_yaml(output_list):
    print(list_as_yaml_file_data(output_list))


def list_as_yaml_file_data(output_list):
    return '\n---\n'.join(map(dump_yaml, output_list))


def dump_yaml(object_to_dump):
    output = yaml.dump(object_to_dump, default_flow_style=False)
    return output.rstrip()


def write_to_file(filename, content):
    with open(filename, 'a') as output_file:
        output_file.write(content)



import copy
import time
from aloft import aws
from aloft import cluster
from aloft import chart_config
from aloft import k8s
from aloft import output
from aloft.process import execute
from tempfile import NamedTemporaryFile


def remove_released_volume_resources(release_id, chart_set, charts, sandbox_name):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    cluster_id = release_config['cluster']
    vpc_id = cluster.get_vpc_id_by_cluster_id(cluster_id)

    for chart in charts:
        bucket_name = get_persistent_volumes_bucket_name(vpc_id)
        filename = get_persistent_volumes_filename(chart_set, release_id, chart, sandbox_name)
        output.print_action(f'Removing released volume resources defined in s3://{bucket_name}/{filename}')
        volume_definitions_filename = save_persistent_volume_definitions_from_s3(bucket_name, filename)

        if volume_definitions_filename:
            execute(f'kubectl delete -f {volume_definitions_filename}')
        else:
            output.print_details(f'No volume resources found.')


def restore_volumes(release_id, chart_set, charts, sandbox_name):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    cluster_id = release_config['cluster']
    vpc_id = cluster.get_vpc_id_by_cluster_id(cluster_id)

    for chart in charts:
        bucket_name = get_persistent_volumes_bucket_name(vpc_id)
        filename = get_persistent_volumes_filename(chart_set, release_id, chart, sandbox_name)
        output.print_action(f'Restoring persistent volumes from s3://{bucket_name}/{filename}')
        volume_definitions_filename = save_persistent_volume_definitions_from_s3(bucket_name, filename)

        if volume_definitions_filename:
            k8s.apply_file(volume_definitions_filename)
        else:
            output.print_details(f'No volume resources found.')


def save_persistent_volume_definitions_from_s3(bucket_name, filename):
    content = aws.get_content_from_s3(bucket_name, filename)
    created_filename = None

    if content:
        with NamedTemporaryFile(mode='w+t', encoding='utf-8', delete=False) as temp_file:
            temp_file.write(content)
            temp_file.write('\n')
            created_filename = temp_file.name

    return created_filename


def unlock_volumes(release_id, chart_set, charts, sandbox_name):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    namespace = release_config['namespace']

    for chart in charts:
        persistent_volume_names = get_chart_persistent_volume_names(chart, namespace)

        if persistent_volume_names:
            unlock_volume(release_id, chart_set, sandbox_name, chart, persistent_volume_names)


def unlock_volume(release_id, chart_set, sandbox_name, chart, persistent_volume_names):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    cluster_id = release_config['cluster']
    vpc_id = cluster.get_vpc_id_by_cluster_id(release_config['cluster'])

    output.print_action(f'Unlocking persistent volumes: {persistent_volume_names}')

    for persistent_volume_name in persistent_volume_names:
        release_persistent_volume(cluster_id, persistent_volume_name)

    bucket_name = get_persistent_volumes_bucket_name(vpc_id)
    filename = get_persistent_volumes_filename(chart_set, release_id, chart, sandbox_name)
    delete_persistent_volume_definitions_from_s3(bucket_name, filename)


def release_persistent_volume(cluster_id, persistent_volume_name):
    persistent_volume = get_persistent_volume(persistent_volume_name)
    set_persistent_volume_reclaim_policy(persistent_volume, 'Delete')
    attach_volume_to_kops(persistent_volume, cluster_id)


def lock_volumes(release_id, chart_set, charts, sandbox_name):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    namespace = release_config['namespace']

    for chart in charts:
        persistent_volume_names = get_chart_persistent_volume_names(chart, namespace)

        if persistent_volume_names:
            lock_volume(release_id, chart_set, sandbox_name, chart, persistent_volume_names)


def lock_volume(release_id, chart_set, sandbox_name, chart, persistent_volume_names):
    release_config = chart_config.get_release_config(chart_set, release_id, sandbox_name)
    cluster_id = release_config['cluster']
    vpc_id = cluster.get_vpc_id_by_cluster_id(release_config['cluster'])
    persistent_volumes = []

    output.print_action(f'Locking persistent volumes: {persistent_volume_names}')

    for persistent_volume_name in persistent_volume_names:
        retain_persistent_volume(cluster_id, persistent_volume_name, persistent_volumes)

    bucket_name = get_persistent_volumes_bucket_name(vpc_id)
    filename = get_persistent_volumes_filename(chart_set, release_id, chart, sandbox_name)
    save_persistent_volume_definitions_to_s3(persistent_volumes, bucket_name, filename)


def retain_persistent_volume(cluster_id, persistent_volume_name, persistent_volumes):
    persistent_volume = get_persistent_volume(persistent_volume_name)
    set_persistent_volume_reclaim_policy(persistent_volume, 'Retain')
    release_volume_from_kops(persistent_volume, cluster_id)
    persistent_volumes.append(clean_persistent_volume_definition(persistent_volume))


def get_persistent_volumes_bucket_name(vpc_id):
    return f'persistent-volumes.{vpc_id}'


def get_persistent_volumes_filename(chart_set, release_id, chart, sandbox_name):
    filename = f'{release_id}_{chart_set}_{chart}.yaml'

    if sandbox_name:
        filename = f'{sandbox_name}_{filename}'

    return filename


def set_persistent_volume_reclaim_policy(persistent_volume, reclaim_policy):
    persistent_volume_name = get_persistent_volume_name(persistent_volume)

    output.print_action(f'Setting persistent volume reclaim policy for {persistent_volume_name} to {reclaim_policy}')

    if persistent_volume['spec']['persistentVolumeReclaimPolicy'] != reclaim_policy:
        reclaim_policy_spec = '\'{"spec":{"persistentVolumeReclaimPolicy":"' + reclaim_policy + '"}}\''
        execute(f'kubectl patch pv {persistent_volume_name} -p {reclaim_policy_spec}')
        persistent_volume['spec']['persistentVolumeReclaimPolicy'] = reclaim_policy


def release_volume_from_kops(persistent_volume, cluster_id):
    persistent_volume_name = get_persistent_volume_name(persistent_volume)
    output.print_action(f'Releasing persistent volume {persistent_volume_name} from Kops.')
    volume_resource_id = get_aws_volume_resource_id(persistent_volume)
    aws.delete_ec2_tags(volume_resource_id,
                        [f'kubernetes.io/cluster/{cluster_id}', 'KubernetesCluster', 'LockedOnCluster'])
    aws.create_ec2_tags(volume_resource_id, {'LockedOnCluster': cluster_id})


def attach_volume_to_kops(persistent_volume, cluster_id):
    persistent_volume_name = get_persistent_volume_name(persistent_volume)
    output.print_action(f'Attaching persistent volume {persistent_volume_name} to Kops.')
    volume_resource_id = get_aws_volume_resource_id(persistent_volume)
    aws.delete_ec2_tags(volume_resource_id,
                        [f'kubernetes.io/cluster/{cluster_id}', 'KubernetesCluster', 'LockedOnCluster'])
    aws.create_ec2_tags(volume_resource_id, {
        f'kubernetes.io/cluster/{cluster_id}': 'owned',
        'KubernetesCluster': cluster_id
    })


def get_aws_volume_resource_id(persistent_volume):
    return persistent_volume['spec']['awsElasticBlockStore']['volumeID'].split('/')[-1]


def get_chart_persistent_volume_names(chart, namespace):
    resource_name = f'{namespace}-{chart}'
    select_by_release_name = [{'key': 'release', 'value': resource_name, 'operator': '='}]
    all_volumes_found = False
    persistent_volume_names = []
    volume_not_found_count = 0

    while not all_volumes_found:
        persistent_volume_claims = k8s.get_resources('pvc', namespace, select_by_release_name)
        missing_volume_names = []

        for persistent_volume_claim in persistent_volume_claims:
            persistent_volume_name = persistent_volume_claim['spec'].get('volumeName')
            if persistent_volume_name:
                persistent_volume_names.append(persistent_volume_name)
            else:
                missing_volume_names.append(persistent_volume_claim['metadata']['name'])

        if len(missing_volume_names) == 0:
            all_volumes_found = True
        else:
            time.sleep(2)
            volume_not_found_count += 1

        if volume_not_found_count > 10:
            raise RuntimeError(f'Unable to retrieve persistent volume names for: {", ".join(missing_volume_names)}')

    return persistent_volume_names


def get_persistent_volume(persistent_volume_name):
    return k8s.get_resource_by_name('pv', persistent_volume_name)


def get_persistent_volume_name(persistent_volume):
    return persistent_volume['metadata']['name']


def clean_persistent_volume_definition(persistent_volume):
    keys_to_keep = ['apiVersion', 'kind', 'metadata', 'spec']
    metadata_keys_to_keep = ['name']
    claim_keys_to_keep = ['apiVersion', 'kind', 'name', 'namespace']

    cleaned_volume = copy.deepcopy(persistent_volume)
    cleaned_volume = {key: cleaned_volume[key] for key in keys_to_keep}
    cleaned_volume['metadata'] = {key: cleaned_volume['metadata'][key] for key in metadata_keys_to_keep}
    cleaned_volume['spec']['claimRef'] = {key: cleaned_volume['spec']['claimRef'][key] for key in claim_keys_to_keep}

    return cleaned_volume


def save_persistent_volume_definitions_to_s3(persistent_volumes, bucket_name, filename):
    output.print_action(f'Saving persistent volume definition to s3://{bucket_name}/{filename}')
    content = output.list_as_yaml_file_data(persistent_volumes)
    output.print_details(content)
    aws.save_content_to_s3(content, bucket_name, filename)


def delete_persistent_volume_definitions_from_s3(bucket_name, filename):
    output.print_action(f'Deleting persistent volume definition from s3://{bucket_name}/{filename}')
    aws.delete_file_in_s3(bucket_name, filename)



from aloft import aws
from aloft import input
from aloft import cluster_config


def get_or_create_vpc(vpc_id):
    vpc_name = get_vpc_name_by_vpc_id(vpc_id)
    vpc = get_vpc(vpc_name)

    if not vpc:
        vpc = create_vpc(vpc_id)

    return vpc


def get_vpc(vpc_name):
    return aws.get_vpc_by_name(vpc_name)


def create_vpc(vpc_id):
    vpc_name = get_vpc_name_by_vpc_id(vpc_id)
    vpc_values_filename = cluster_config.get_vpc_config_filename(vpc_id)
    vpc_config = input.read_yaml_file(vpc_values_filename)

    vpc = aws.create_vpc(vpc_name, vpc_config['networkCIDR'])
    aws.create_and_attach_internet_gateway(vpc)
    aws.create_and_attach_vpn_gateway(vpc)
    return vpc


def delete_vpc(vpc_id):
    vpc_name = get_vpc_name_by_vpc_id(vpc_id)
    aws.delete_network_and_vpc(vpc_name)


def get_vpc_config(vpc_id):
    vpc_values_filename = cluster_config.get_vpc_config_filename(vpc_id)
    return input.read_yaml_file(vpc_values_filename)


def get_vpc_name_by_vpc_id(vpc_id):
    return vpc_id.split('.', 1)[0]



from setuptools import setup

setup(name='aloft',
      version='0.4.4',
      description='Tool to manage Kubernetes clusters and helm charts across multiple AWS accounts and clusters',
      packages=['aloft'],
      python_requires='>=3.6',
      scripts=['bin/aloft'],
      zip_safe=False,
      install_requires=['botocore>=1.12.6', 'boto3>=1.9.6', 'jinja2>=2.10', 'PyYAML>=3.12', 'docopt>=0.6.2']
      )



import os
from aloft import config
from aloft import input
from tempfile import NamedTemporaryFile


def get_release_config(chart_set, release_id, sandbox_name):
    release_configs = get_releases_config(chart_set)
    release_config = release_configs[release_id]

    if sandbox_name:
        sandbox_config = release_configs['sandbox']
        sandbox_config['valuesPath'] = release_config['valuesPath']
        sandbox_config['namespace'] += f'-{sandbox_name}'
        release_config = sandbox_config

    add_sandbox_properties(release_config, sandbox_name)

    return release_config


def add_sandbox_properties(release_config, sandbox_name):
    if sandbox_name:
        sandbox_name_prefixed_with_dash = f'-{sandbox_name}'
        sandbox_name_suffixed_with_dash = f'{sandbox_name}-'
    else:
        sandbox_name = ''
        sandbox_name_prefixed_with_dash = ''
        sandbox_name_suffixed_with_dash = ''

    if 'properties' not in release_config:
        release_config['properties'] = {}

    release_config['properties']['sandboxName'] = sandbox_name
    release_config['properties']['sandboxNamePrefixedWithDash'] = sandbox_name_prefixed_with_dash
    release_config['properties']['sandboxNameSuffixedWithDash'] = sandbox_name_suffixed_with_dash


def get_charts_to_use(chart_set):
    charts_config = get_charts_config(chart_set)
    charts_to_apply = []

    for chart in charts_config['charts']:
        charts_to_apply.append(chart['name'])

    return charts_to_apply


def get_releases_config(chart_set):
    chart_set_config_directory = get_chart_set_config_directory(chart_set)
    return input.read_yaml_file(f'{chart_set_config_directory}/releases.yaml')


def get_chart_config_for_chart(chart_set, chart):
    charts_config = get_charts_config(chart_set)
    filtered_charts = list(filter(lambda x: x['name'] == chart, charts_config['charts']))
    chart_config = None

    if len(filtered_charts) == 1:
        chart_config = filtered_charts[0]

    return chart_config


def get_charts_config(chart_set):
    chart_set_config_directory = get_chart_set_config_directory(chart_set)
    return input.read_yaml_file(f'{chart_set_config_directory}/charts.yaml')


def get_secrets_config_for_chart(chart_set, chart):
    secrets_config = get_secrets_config(chart_set)
    secrets_config_for_chart = None

    if secrets_config:
        secrets_config_for_chart = secrets_config.get(chart, None)

    return secrets_config_for_chart


def get_secrets_config(chart_set):
    chart_set_config_directory = get_chart_set_config_directory(chart_set)
    return input.read_yaml_file(f'{chart_set_config_directory}/secrets.yaml')


def get_chart_set_config_directory(chart_set):
    base_chart_config_directory = get_base_chart_config_directory()
    return f'{base_chart_config_directory}/chart-sets/{chart_set}'


def get_base_chart_config_directory():
    return config.get_aloft_config_directory()


def generate_value_files(release_id, chart_set, chart, release_config):
    values_config_directory = get_chart_set_config_directory(chart_set)
    value_paths = release_config.get('valuesPath', f'values/{release_id}').split('/')
    properties = release_config['properties']
    value_filenames = []

    for path in value_paths:
        values_config_directory += f'/{path}'
        append_generated_values_file(values_config_directory, chart, properties, value_filenames)

    return value_filenames


def append_generated_values_file(values_config_directory, chart, properties, value_filenames):
    filename = f'{values_config_directory}/{chart}.yaml'

    if os.path.isfile(filename):
        rendered_values = input.render_template_file(filename, properties)
        with NamedTemporaryFile(mode='w+t', encoding='utf-8', delete=False) as temp_file:
            temp_file.write(rendered_values)
            temp_file.write('\n')
            value_filenames.append(temp_file.name)
